from urlmagic.views import filtered


class MyMixin(object):
    additional_kwargs = {
        "%s_owner" % filtered.FilteredViewMixin.REQUEST_USER: True,
    }


class MyListView(MyMixin, filtered.FilteredListView):
    pass


class MyDetailView(MyMixin, filtered.FilteredDetailView):
    pass


class MyCreateView(MyMixin, filtered.FilteredCreateView):
    pass


class MyUpdateView(MyMixin, filtered.FilteredUpdateView):
    pass


class MyDeleteView(MyMixin, filtered.FilteredDeleteView):
    pass



from urlmagic import mixins
from django.http import Http404
from django.core.exceptions import PermissionDenied
from django.contrib.admin.util import NotRelationField
from django.contrib.admin.util import get_model_from_relation
from django.db.models.constants import LOOKUP_SEP
from urlmagic.views import core


# Regular views can be statically filtered by providing a modified queryset paramter (e.g. Foo.objects.filter(flag=True)).
# Sometimes it becomes necessary to filter dynamically using the URL as an indicator of the filtered context object.
# Consider examples such as the following:
#
# /alphas/          Alpha.objects.all()
# /alphas/active/   Alpha.objects.filter(active=True)
# /alphas/1/        Alpha.objects.get(pk=1)
# /alphas/1/betas/  Beta.objects.filter(alphas__pk=1)
# /alphas/1/beta/   Beta.objects.get(alphas__pk=1)
#
# It's the concept in these last two examples that these filtered views address.


class FilteredViewMixin(object):
    context_slug = None
    REQUEST_USER = "request_user"
    additional_kwargs = {}

    def trim_request_user(self, string):
        return string[len(self.REQUEST_USER) + 1:] if string.startswith(self.REQUEST_USER) else string

    def get_merged_kwargs(self):
        kwargs = {}
        kwargs.update(self.kwargs)
        kwargs.update(self.additional_kwargs)
        return kwargs

    def get_filter_kwargs(self):
        if not self.request.user.is_authenticated():
            for key in self.kwargs.keys():
                if key.startswith(self.REQUEST_USER):
                    raise PermissionDenied
        return dict(
            (
                self.trim_request_user(key),
                self.request.user if key.startswith(self.REQUEST_USER) else value,
            )
            for key, value in self.get_merged_kwargs().iteritems()
        )

    def get_queryset(self):
        return super(FilteredViewMixin, self).get_queryset().filter(**self.get_filter_kwargs())

    def get_context_data(self, *args, **kwargs):
        context = super(FilteredViewMixin, self).get_context_data(*args, **kwargs)

        # Add in the queryset used for this filtered view
        context.update({"view_filter_kwargs": self.get_filter_kwargs()})

        # Add in the object that's been filtered on if there's only one or if it's been explicitly provided
        if self.context_slug or len(self.get_merged_kwargs()) == 1:
            context_pointer = self.model
            context_slug = self.context_slug or self.get_merged_kwargs().keys()[0]
            field_names = context_slug.split(LOOKUP_SEP)
            try:
                for field_name in field_names:
                    field_name = self.trim_request_user(field_name)
                    if field_name == "pk":
                        field_name = context_pointer._meta.pk.name
                    context_pointer = get_model_from_relation(context_pointer._meta.get_field_by_name(field_name)[0])
                field_name = "pk"
                field_value = context["view_filter_kwargs"][self.trim_request_user(context_slug)].pk
            except NotRelationField:
                field_name = field_names[-1]
                field_value = context["view_filter_kwargs"][context_slug]
            context["context_object"] = context_pointer.objects.get(**{field_name: field_value})

        return context


class FilteredListView(FilteredViewMixin, core.ContextListView):
    pass


class FilteredDetailView(FilteredViewMixin, core.ContextDetailView):
    pass


class FilteredCreateView(FilteredViewMixin, mixins.FormRequestViewMixin, core.ContextCreateView):
    pass


class FilteredUpdateView(FilteredViewMixin, mixins.FormRequestViewMixin, core.ContextUpdateView):
    pass


class FilteredDeleteView(FilteredViewMixin, core.ContextDeleteView):
    pass



from django.contrib import messages
from django.http import Http404
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from urlmagic.utils import get_user_field_names


class ContextViewMixin(object):
    extra_context = {}

    def get_context_data(self, *args, **kwargs):
        context = super(ContextViewMixin, self).get_context_data(*args, **kwargs)
        context.update(self.extra_context)
        return context


class ObjectAlreadyExists(Exception):
    pass


class RedirectOnExistsMixin(object):
    redirect_on_exists = None
    def get(self, request, *args, **kwargs):
        try:
            return super(RedirectOnExistsMixin, self).get(request, *args, **kwargs)
        except ObjectAlreadyExists:
            if self.redirect_on_exists:
                # TODO: Investigate why django.shortcuts.redirect isn't properly handling relative URLs like "../add/"
                if self.redirect_on_exists.startswith("."):
                    return HttpResponseRedirect(self.redirect_on_exists)
                else:
                    return redirect(self.redirect_on_exists)
            else:
                raise


class RedirectOn404Mixin(object):
    redirect_on_404 = None

    def get(self, request, *args, **kwargs):
        try:
            return super(RedirectOn404Mixin, self).get(request, *args, **kwargs)
        except Http404:
            if self.redirect_on_404:
                # TODO: Investigate why django.shortcuts.redirect isn't properly handling relative URLs like "../add/"
                if self.redirect_on_404.startswith("."):
                    return HttpResponseRedirect(self.redirect_on_404)
                else:
                    return redirect(self.redirect_on_404)
            else:
                raise


class ChokeIfAlreadyExists(object):
    request = None
    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super(ChokeIfAlreadyExists, self).__init__(*args, **kwargs)
        if self.instance._meta.model.objects.filter(**dict(
            (field_name, self.request.user)
            for field_name in getattr(self, "user_fields", get_user_field_names(self.instance))
        )).exists():
            raise ObjectAlreadyExists


class AutomaticUserFormMixin(object):
    def save(self, *args, **kwargs):
        for field_name in getattr(self, "user_fields", get_user_field_names(self.instance)):
            setattr(self.instance, field_name, self.request.user)
        return super(AutomaticUserFormMixin, self).save(*args, **kwargs)


class SingularViewMixin(object):
    def adjust_kwargs(self):
        pk = self.kwargs.get(self.pk_url_kwarg, None)
        slug = self.kwargs.get(self.slug_url_kwarg, None)
        if pk is None and slug is None:
            queryset = self.get_queryset()
            if queryset.count():
                self.kwargs["pk"] = queryset.first().pk
            else:
                raise Http404

    def get(self, request, *args, **kwargs):
        self.adjust_kwargs()
        return super(SingularViewMixin, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.adjust_kwargs()
        return super(SingularViewMixin, self).post(request, *args, **kwargs)


class FormRequestViewMixin(object):
    def get_form_kwargs(self):
        kwargs = super(FormRequestViewMixin, self).get_form_kwargs()
        if hasattr(self.get_form_class(), "request"):
            kwargs["request"] = self.request
        return kwargs

    def get_form(self, form_class):
        form = super(FormRequestViewMixin, self).get_form(form_class)
        form.request = self.request
        return form


class SuccessRedirectMixin(object):
    def get_success_url(self, *args, **kwargs):
        if self.request.REQUEST.get("success_url", None):
            return self.request.REQUEST.get("success_url")
        else:
            return super(SuccessRedirectMixin, self).get_success_url(*args, **kwargs)


class MessageSaveMixin(object):
    success_message = "Save successful."

    def form_valid(self, *args, **kwargs):
        result = super(MessageSaveMixin, self).form_valid(*args, **kwargs)
        messages.success(self.request, self.success_message)
        return result


class MessageDeleteMixin(object):
    success_message = "Deletion successful."

    def delete(self, *args, **kwargs):
        result = super(MessageDeleteMixin, self).delete(*args, **kwargs)
        messages.success(self.request, self.success_message)
        return result



from django.views.generic.detail import DetailView
from django.views.generic.edit import UpdateView, CreateView, DeleteView
from django.views.generic.list import ListView
from urlmagic import mixins


class ContextListView(mixins.ContextViewMixin, ListView):
    pass


class ContextDetailView(mixins.ContextViewMixin, mixins.RedirectOn404Mixin, mixins.SingularViewMixin, DetailView):
    pass


class ContextCreateView(mixins.ContextViewMixin, mixins.RedirectOnExistsMixin, mixins.SuccessRedirectMixin, mixins.MessageSaveMixin, CreateView):
    pass


class ContextUpdateView(mixins.ContextViewMixin, mixins.RedirectOn404Mixin, mixins.SuccessRedirectMixin, mixins.MessageSaveMixin, mixins.SingularViewMixin, UpdateView):
    pass


class ContextDeleteView(mixins.ContextViewMixin, mixins.RedirectOn404Mixin, mixins.SuccessRedirectMixin, mixins.MessageDeleteMixin, mixins.SingularViewMixin, DeleteView):
    pass
